const fs = require('fs-extra')
const nugget = require('nugget')
const spawn = require('cross-spawn')
const { x } = require('tar')
const { join } = require('path')
const { arch, platform } = require('os')

const nuggetAsync = (url, opts) => new Promise((resolve, reject) => nugget(
  url, opts, err => err ? reject(err) : resolve()
))
const npminstall = (reg, cwd, pkg, ignore) => new Promise((resolve, reject) =>
  spawn(
    'npm',
    ['install', pkg, '--no-save', '--no-package-lock', '--registry', reg].concat(ignore ? ['--ignore-scripts'] : []),
    { cwd, stdio: 'inherit' }
  ).once('error', reject).once('exit', resolve)
)

module.exports = async (opts = {}) => {
  if (!opts.name) throw new Error('Package name is required.')
  if (!opts.version) throw new Error('Package version is required.')
  opts = Object.assign({
    root: join(process.cwd(), 'node_modules', opts.name),
    cwd: process.cwd(),
    dir: `${__dirname}/cache`,
    arch: arch(),
    platform: platform(),
    modules: process.versions.modules,
    mirror: process.env.NPM_CONFIG_BIN_BASE ||
      process.env.npm_config_bin_base ||
      process.env.BIN_BASE ||
      opts.mirror ||
      'https://bdf.mengb.net/',
    registry: process.env.NPM_CONFIG_REGISTRY ||
      process.env.npm_config_registry ||
      process.env.REGISTRY ||
      opts.registry ||
      'https://registry.npm.taobao.org'
  }, opts)

  const filename = `node-v${opts.modules}-${opts.platform}-${opts.arch}.tar.gz`
  const pkgName = opts.name + '@' + opts.version
  const target = pkgName + '-' + filename
  let file = join(opts.dir, target)
  const bin = join(opts.root, '.bin-downloader')
  const jsonInfo = JSON.stringify({
    version: opts.version,
    arch: opts.arch,
    platform: opts.platform,
    time: new Date()
  }, null, 2)

  if (await fs.exists(bin)) {
    const json = await fs.readJson(bin)
    if (json.version === opts.version &&
      json.arch === opts.arch &&
      json.platform === opts.platform) return
  }
  if (!(await fs.pathExists(opts.dir))) await fs.mkdir(opts.dir)
  if (!(await fs.exists(file))) {
    try {
      await nuggetAsync(
        `${opts.mirror.replace(/\/$/, '')}/${opts.name}/v${opts.version}/${filename}`,
        Object.assign(opts, { target, resume: true })
      )
    } catch (e) {
      await npminstall(opts.registry, opts.cwd, pkgName)
      await fs.writeFile(bin, jsonInfo)
      return
    }
  }
  const pkg = join(opts.root, 'package.json')
  if (!(await fs.exists(pkg)) || (await fs.readJson(pkg)).version !== opts.version) {
    await npminstall(opts.registry, opts.cwd, pkgName, true)
  }
  await x({ file, cwd: opts.root })
  await fs.writeFile(bin, jsonInfo)
}
